"""
    A simple, low level, ngas retriever with verification.

    Given a file, and a location, interact with NGAS to find the server, and initiate a direct copy from NGAS
    and verify that the transfer was completed appropriately.

"""
import os
import sys
import requests
import argparse as ap
import logging
import logging.handlers

from ._version import ___version___ as version
from pymygdala import LogHandler
from pycapo import CapoConfig
from bs4 import BeautifulSoup

# Constants:  Might want to become CAPO properties.
_NGAS_MASTER_NODE = 'aocngas-master'
_NGAS_PORT = 7777

_DESCRIPTION = """AAT/PPI uitility for low-level NGAS interactions. Version {}
                This tool interacts with the NGAS HTTP interface to facilitate the transfer of a file to disk."""

# set up logging:
logger = logging.getLogger("ngas_retriever")
logger.setLevel(logging.DEBUG)


def _make_parser():
    r""" Build a command line and/or argument parser """

    result = ap.ArgumentParser(description=_DESCRIPTION.format(version),
                               formatter_class=ap.RawTextHelpFormatter)
    result.add_argument('-P', '--profile', action='store', default='',
                        help='profile name to use, e.g. test, production')
    result.add_argument('-d', '--destination', type=str, action='store', default='',
                        help='Location to place the file (default is current working directory)')
    # result.add_argument('-e', '--email', type=str, action='store', default='', help='Email for completion notification')
    result.add_argument('filename', type=str, action='store',
                        help='NGAS file id to retrieve.')
    return result

# A typical response we need to extract information from:
#
# <?xml version="1.0" encoding="UTF-8"?>
# <!DOCTYPE NgamsStatus SYSTEM "http://aocngas-master.aoc.nrao.edu:7777/RETRIEVE?internal=ngamsStatus.dtd">
# <NgamsStatus>
#   <Status CompletionTime="2018-10-24T11:27:24.637" Date="2018-10-24T11:27:24.629" HostId="aocngas-master" Message="Successfully handled command STATUS" RequestId="4150" RequestTime="2018-10-24T11:27:24.623" State="ONLINE" Status="SUCCESS" SubState="IDLE" Version="v4.2.2-ALMA/2013-12-101T16:00:00"/>
#   <DiskStatus Archive="NRAO-ARCHIVE" AvailableMb="83" BytesStored="1091949384027" Checksum="" Completed="1" CompletionDate="2010-08-02T09:17:17.368" DiskId="3c790665f4f353cb2ae19faaa85ea754" HostId="aocngas-4" InstallationDate="2009-12-04T13:23:21.957" LastCheck="" LogicalName="NRAO-ARCH-M-000184" Manufacturer="Western Digital" MountPoint="/NGAS/volumes/Volume01" Mounted="1" NumberOfFiles="33445" SlotId="Volume01" TotalDiskWriteTime="19938" Type="raid1_part2">
#     <FileStatus AccessDate="" Checksum="711574960" ChecksumPlugIn="ngamsGenCrc32" Compression="" CreationDate="2010-02-11T10:35:02.000" FileId="uid___evla_sdm_X1265909221067.sdm" FileName="data/2010-02-11/1/uid___evla_sdm_X1265909221067.sdm" FileSize="395" FileStatus="00000000" FileVersion="1" Format="application/x-sdm" Group="" Ignore="0" IngestionDate="2010-02-11T10:35:02.032" ModificationDate="" Owner="" Permissions="" Tag="" UncompressedFileSize="395"/>
#   </DiskStatus>
# </NgamsStatus>

def perform_preliminary_lookup(filename):

    ## Eventually, we'll want the master node & ports as CAPO properties: config = CapoConfig()

    status_url = "http://{}.aoc.nrao.edu:{}/STATUS?file_id={}".format(_NGAS_MASTER_NODE, _NGAS_PORT, filename)

    status_info = requests.get(status_url)

    if requests.codes.ok == status_info.status_code:

        # read and parse out the HostId, FileSize (maybe FileStatus?)
        xml_content = status_info.text
        response_handle = BeautifulSoup(xml_content, "lxml-xml")
        # print(xml_content)

        disk_handle = response_handle.NgamsStatus.DiskStatus
        hostname = disk_handle.get("HostId")


        file_handle = response_handle.find("FileStatus")

        file_status = file_handle.get('FileStatus')
        file_size   = file_handle.get('FileSize')

        return hostname, file_size

    logger.error(f"Got code {status_info.status_code} from NGAS, cannot proceed.")

    # If something went wrong, indicate that by the return value
    return None, None


def retrieve_file(filename, server, destination, file_size):
    #
    # This is a pure 'yank and put', no renaming or anything fancy.
    #

    retrieve_url = "http://{}.aoc.nrao.edu:{}/RETRIEVE".format(server, _NGAS_PORT)

    # The details of the retrieval request
    parameters = {
        'file_id':filename,
        'processing':'ngamsDirectCopyDppi',
        'processingPars':'outfile='+destination+'/'+filename
    }

    copy_response = requests.get(retrieve_url, params=parameters)

    if requests.codes.ok == copy_response.status_code:

        # xml_content = copy_response.text
        # response_handle = BeautifulSoup(xml_content, "lxml-xml")

        logger.info("Success!")

        # Unfortunately, the return from the direct copy call doesn't provide us much:
        #
        # {'Server': 'NGAMS/v4.2.2-ALMA/2013-12-101T16:00:00', 'Date': 'Wed, 24 Oct 2018 21:44:43 GMT', 'Expires': 'Wed, 24 Oct 2018 21:44:43 GMT', 'Content-type': 'application/x-bdf', 'Content-length': '74', 'Content-disposition': 'attachment; filename=""', 'Accept-Ranges': 'bytes'}
        # /NGAS/volumes/Volume32/data/2018-10-23/1/uid____evla_bdf_1540297250925.bdf
    else:
        logger.error(f"Direct Copy Call failed with code: {copy_response.status_code}")

        xml_content = copy_response.text
        response_handle = BeautifulSoup(xml_content, "lxml-xml")

        status_handle = response_handle.NgamsStatus.Status
        error_message = status_handle.get("Message")
        logger.error(f"Error details: {error_message}")

    logger.info(f"{filename} retrieval complete.")


def ngas_retriever():
    """The heart of the retrieval process """

    parser = _make_parser()

    arguments = parser.parse_args()

    # Shamelessly stolen from epilogue with a twist: allow for explict profile setting via the CL
    if 'CAPO_PROFILE' not in os.environ and '' == arguments.profile:
        # try to synthesize a profile from our installation root
        profile = os.path.abspath(sys.argv[0]).split(os.path.sep)[-3]
        os.environ['CAPO_PROFILE'] = profile
        print('No CAPO_PROFILE explicitly set, synthesizing {0} by path'.format(str(profile)))
    elif '' != arguments.profile:
        os.environ['CAPO_PROFILE'] = arguments.profile

    # integrate the logs with pymygdala's logging system
    handler = LogHandler(profile=arguments.profile, application='aoc_ngas_retriever')
    handler.setLevel(logging.DEBUG)
    logger.addHandler(handler)


    if '' == arguments.destination:
        delivery_location = os.getcwd()
    else:
        delivery_location = arguments.destination

    if delivery_location.endswith('/'):
        # remove any trailing / for consistency:
        delivery_location = delivery_location[:-1]

    server, file_size = perform_preliminary_lookup(arguments.filename)

    # If we don't have the info we need, exit with an error:
    if (None == server) or (None == file_size):
        exit(-10)

    logger.info(f"We want to get {arguments.filename} from {server}, and we expect {file_size} bytes")

    retrieve_file(arguments.filename, server, delivery_location, file_size)

if __name__ == "__main__":
    ngas_retriever()